import math
import numpy as np
import matplotlib.pyplot as plt

import matplotlib.pyplot as plt
import numpy as np
import torch

CREATE_MOVIE = 0


def r2w(z_r, x):
    z_w = torch.zeros(3)
    R = torch.vstack((torch.hstack((torch.cos(x[2]), -torch.sin(x[2]))), torch.hstack((torch.sin(x[2]), torch.cos(x[2])))))
    z_w[0:2] = R @ z_r[0:2] + x[0:2]
    z_w[2] = z_r[2] + x[2]
    return z_w

def spring(start, end, nodes, width):
    """!
    Return a list of points corresponding to a spring.

    @param r1 (array-like) The (x, y) coordinates of the first endpoint.
    @param r2 (array-like) The (x, y) coordinates of the second endpoint.
    @param nodes (int) The number of spring "nodes" or coils.
    @param width (int or float) The diameter of the spring.
    @return An array of x coordinates and an array of y coordinates.
    """

    # Check that nodes is at least 1.
    nodes = max(int(nodes), 1)

    # Convert to numpy array to account for inputs of different types/shapes.
    start, end = np.array(start).reshape((2,)), np.array(end).reshape((2,))

    # If both points are coincident, return the x and y coords of one of them.
    if (start == end).all():
        return start[0], start[1]

    # Calculate length of spring (distance between endpoints).
    length = np.linalg.norm(np.subtract(end, start))

    # Calculate unit vectors tangent (u_t) and normal (u_t) to spring.
    u_t = np.subtract(end, start) / length
    u_n = np.array([[0, -1], [1, 0]]).dot(u_t)

    # Initialize array of x (row 0) and y (row 1) coords of the nodes+2 points.
    spring_coords = np.zeros((2, nodes + 2))
    spring_coords[:,0], spring_coords[:,-1] = start, end

    # Check that length is not greater than the total length the spring
    # can extend (otherwise, math domain error will result), and compute the
    # normal distance from the centerline of the spring.
    normal_dist = math.sqrt(max(0, width**2 - (length**2 / nodes**2))) / 2

    # Compute the coordinates of each point (each node).
    for i in range(1, nodes + 1):
        spring_coords[:,i] = (
            start
            + ((length * (2 * i - 1) * u_t) / (2 * nodes))
            + (normal_dist * (-1)**i * u_n))

    return spring_coords[0,:], spring_coords[1,:]

def plot_cf(x1_np, length=1, color1='red', color2='blue'):
    ax = plt.gca()
    ax.quiver(x1_np[0], x1_np[1], length*np.cos(x1_np[2]), length*np.sin(x1_np[2]), angles='xy', scale_units='xy', scale=1, color=color1, width=0.01 +0.003*length, linewidth=1, edgecolor='w', zorder=10)
    ax.quiver(x1_np[0], x1_np[1], -length*np.sin(x1_np[2]), length*np.cos(x1_np[2]), angles='xy', scale_units='xy', scale=1, color=color2, width=0.01+0.003*length, linewidth=1, edgecolor='w', zorder=10)

def plot_cam(x1_np, length=1, color1='red', color2='blue'):
    ax = plt.gca()
    ax.quiver(x1_np[0], x1_np[1], length*np.cos(x1_np[2]), length*np.sin(x1_np[2]), angles='xy', scale_units='xy', scale=1, color=color1, width=0.01 +0.003*length, linewidth=1, edgecolor='w', zorder=10)
    ax.quiver(x1_np[0], x1_np[1], -length*np.sin(x1_np[2]), length*np.cos(x1_np[2]), angles='xy', scale_units='xy', scale=1, color=color2, width=0.01+0.003*length, linewidth=1, edgecolor='w', zorder=10)
    a = r2w(torch.tensor([1.0, 1.5, 0.0]), torch.tensor(x1_np))
    b = r2w(torch.tensor([1.0, -1.5, 0.0]), torch.tensor(x1_np))

    ax.plot([a[0], b[0]], [a[1], b[1]], color='black', markersize=5, linewidth=3, mew=5, zorder=14)
