import matplotlib.pyplot as plt
import numpy as np
import torch
from drw_tools import spring


z1_gps, s1_gps = torch.tensor([2.0, 4.0]), torch.tensor([[1.0, 0.0], [1.0, 0.0]])
z2_gps, s2_gps = torch.tensor([6.0, 5.0]), torch.tensor([[1.0, 0.0], [0.0, 1.0]])

z12_odom, s12_odom = torch.tensor([4.0, 1.0]), torch.tensor([[1.0, 0.0], [0.0, 1.0]])
x1 = torch.tensor([4.0, 1.0], requires_grad=True)       # state: x,y position in time=1
x2 = torch.tensor([5.0, 8.0], requires_grad=True)        # state: x,y position in time=2
omega = torch.tensor([np.pi * 0.4], requires_grad=True)  # control: shooting angle
v = torch.tensor([15.0], requires_grad=True)             # control: shooting velocity
g = torch.tensor([9.81], requires_grad=True)             # gravitation acceleration
dt = torch.tensor([1.0], requires_grad=True)             # seconds between time instances

opt = torch.optim.Adam([x1, x2], lr=0.5)

fig, ax = plt.subplots()

for i in range(200):
    g12 = torch.cat((v * dt * torch.cos(omega), v * dt * torch.sin(omega) - 0.5 * g * dt *dt))
    loss = ...  # <------ TODO FILL IN
    # Hint1: You want to make "x1" as close as possible to "z1_gps"
    #        You want to make "x1+g12" as close as possible to "x2", etc...
    # Hint2: In order to compute squared distance between vectors a, b
    #        the following notation can be used in pytorch:
    #        ((a-b)**2).sum()
    loss.backward()
    opt.step()
    opt.zero_grad()

    print('loss=', loss.detach().numpy())


    # visu
    z1_gps_np, z2_gps_np, z12_odom_np = z1_gps.detach().numpy(), z2_gps.detach().numpy(), z12_odom.detach().numpy()
    x1_np, x2_np, theta_np, v_np, g_np = x1.detach().numpy(), x2.detach().numpy(), omega.detach().numpy(), v.detach().numpy(), g.detach().numpy()
    ax.clear()
    plt.grid()
    ax.plot(*spring(x1_np, z1_gps_np, 22, 1.2), c="black")
    ax.plot(*spring(x2_np, z2_gps_np, 22, 1.2), c="black")
    ax.plot(*spring(x1_np+z12_odom_np, x2_np, 22, 1.2), c="blue")
    ax.plot(*spring(x1_np+g12.detach().numpy(), x2_np, 22, 1.2), c="green")
    t = np.linspace(0, dt.detach().numpy(), 11)
    ax.plot(x1_np[0] + v_np*t*np.cos(theta_np), x1_np[1] + v_np*t*np.sin(theta_np) - 0.5*g_np*(t**2), 'kx-', markersize=3, mew=3)
    ax.quiver(x1_np[0], x1_np[1], z12_odom_np[0], z12_odom_np[1], angles='xy', scale_units='xy', scale=1, color='k', width=0.02, linewidth=1, edgecolor='w', zorder=10)

    ax.plot(z1_gps_np[0], z1_gps_np[1], 'rx', markersize=10, mew=3)
    ax.plot(z2_gps_np[0], z2_gps_np[1], 'rx', markersize=10, mew=3)
    ax.plot(x1_np[0], x1_np[1], 'wo', markersize=14, mew=3)
    ax.plot(x1_np[0], x1_np[1], 'go', markersize=10, mew=3)
    ax.plot(x2_np[0], x2_np[1], 'wo', markersize=14, mew=3)
    ax.plot(x2_np[0], x2_np[1], 'go', markersize=10, mew=3)

    ax.set_aspect("equal", "box")

    plt.xlim(0, 10)
    plt.ylim(0, 10)
    plt.pause(0.001)
