import matplotlib.pyplot as plt
import numpy as np
import torch
from drw_tools import spring, plot_cf
import os



def r2w(z_r, x):
    z_w = torch.zeros(3)
    R = torch.vstack((torch.hstack((torch.cos(x[2]), -torch.sin(x[2]))), torch.hstack((torch.sin(x[2]), torch.cos(x[2])))))
    z_w[0:2] = R @ z_r[0:2] + x[0:2]
    z_w[2] = z_r[2] + x[2]
    return z_w


def w2r(z_w, x):
    z_r = torch.zeros(3)
    R = torch.vstack((torch.hstack((torch.cos(x[2]), -torch.sin(x[2]))), torch.hstack((torch.sin(x[2]), torch.cos(x[2])))))
    z_r[0:2] = R.t() @ (z_w[0:2] - x[0:2])
    z_r[2] = z_w[2] - x[2]
    return z_r


fig, ax = plt.subplots()
x1 = torch.tensor([2.0, 4.0, 0.0])  # ground truth
x2 = torch.tensor([7.0, 2.0, np.pi / 4])  # ground truth

m1 = torch.tensor([5.0, 6.0, 0.0])
m2 = torch.tensor([2.0, 2.0, 0.0])
m3 = torch.tensor([8.0, 5.0, 0.0])

z1_m1, z1_m2, z1_m3 = w2r(m1, x1), w2r(m2, x1), w2r(m3, x1)
z2_m1, z2_m2, z2_m3 = w2r(m1, x2), w2r(m2, x2), w2r(m3, x2)
z12_odom = torch.tensor([5.0, -2.0, np.pi / 4])

x1 = torch.tensor([8.0, 3.0, 0.0], requires_grad=True)  # ini
x2 = torch.tensor([4.0, 6.0, np.pi / 4], requires_grad=True)  # ini

m1 = torch.tensor([4.0, 6.0, 0.0],  requires_grad=True)
m2 = torch.tensor([7.0, 8.0, 0.0],  requires_grad=True)
m3 = torch.tensor([8.0, 7.0, 0.0],  requires_grad=True)


opt = torch.optim.Adam([x1, x2, m1, m2, m3], lr=0.1)
for i in range(150):
    loss = ...  # <------ TODO FILL IN
    # Hint1: You want to make "w2r(m1, x1)" as close as possible to "z1_m1"
    #        You want to make "w2r(x2, x1)" as close as possible to "z12_odom", etc...
    # Hint2: In order to compute squared distance between vectors a, b
    #        the following notation can be used in pytorch:
    #        ((a-b)**2).sum()
    #
    loss.backward()
    opt.step()
    opt.zero_grad()

    print('loss=', loss.detach().numpy())

    # visu
    m1_np, m2_np, m3_np = m1.detach().numpy(), m2.detach().numpy(), m3.detach().numpy()
    x1_np, x2_np = x1.detach().numpy(), x2.detach().numpy()
    z12_odom_np = r2w(z12_odom, x1).detach().numpy()
    zw1_m1_np, zw1_m2_np, zw1_m3_np = r2w(z1_m1, x1).detach().numpy(), r2w(z1_m2, x1).detach().numpy(), r2w(z1_m3, x1).detach().numpy()
    zw2_m1_np, zw2_m2_np, zw2_m3_np = r2w(z2_m1, x2).detach().numpy(), r2w(z2_m2, x2).detach().numpy(), r2w(z2_m3, x2).detach().numpy()
    ax.clear()
    plt.grid()

    ax.plot(*spring(m1_np[0:2], zw1_m1_np[0:2], 22, 0.7), c="black")
    ax.plot(*spring(m2_np[0:2], zw1_m2_np[0:2], 22, 0.7), c="black")
    ax.plot(*spring(m3_np[0:2], zw1_m3_np[0:2], 22, 0.7), c="black")

    ax.plot(*spring(m1_np[0:2], zw2_m1_np[0:2], 22, 0.7), c="black")
    ax.plot(*spring(m2_np[0:2], zw2_m2_np[0:2], 22, 0.7), c="black")
    ax.plot(*spring(m3_np[0:2], zw2_m3_np[0:2], 22, 0.7), c="black")

    ax.plot(m1_np[0], m1_np[1], 'o', color='lime', markersize=10, mew=3, zorder=12)
    ax.plot(m2_np[0], m2_np[1], 'o', color='lime', markersize=10, mew=3, zorder=12)
    ax.plot(m3_np[0], m3_np[1], 'o', color='lime', markersize=10, mew=3, zorder=12)
    plot_cf(m1_np, 1, 'red', 'blue')
    plot_cf(m2_np, 1, 'red', 'blue')
    plot_cf(m3_np, 1, 'red', 'blue')

    ax.plot([x1_np[0], zw1_m1_np[0]], [x1_np[1], zw1_m1_np[1]], '--o', color='gray', markersize=10, mew=3, zorder=12)
    ax.plot([x1_np[0], zw1_m2_np[0]], [x1_np[1], zw1_m2_np[1]], '--o', color='gray', markersize=10, mew=3, zorder=12)
    ax.plot([x1_np[0], zw1_m3_np[0]], [x1_np[1], zw1_m3_np[1]], '--o', color='gray', markersize=10, mew=3, zorder=12)
    plot_cf(zw1_m1_np, 1)
    plot_cf(zw1_m2_np, 1)
    plot_cf(zw1_m3_np, 1)

    ax.plot([x2_np[0], zw2_m1_np[0]], [x2_np[1], zw2_m1_np[1]], '--o', color='gray', markersize=10, mew=3)
    ax.plot([x2_np[0], zw2_m2_np[0]], [x2_np[1], zw2_m2_np[1]], '--o', color='gray', markersize=10, mew=3)
    ax.plot([x2_np[0], zw2_m3_np[0]], [x2_np[1], zw2_m3_np[1]], '--o', color='gray', markersize=10, mew=3)
    plot_cf(zw2_m1_np, 1)
    plot_cf(zw2_m2_np, 1)
    plot_cf(zw2_m3_np, 1)

    ax.quiver(x1_np[0], x1_np[1], z12_odom_np[0] - x1_np[0], z12_odom_np[1] - x1_np[1], angles='xy', scale_units='xy', scale=1, color='k', width=0.02, linewidth=1, edgecolor='w', zorder=10)

    ax.plot(*spring(x2_np[0:2], z12_odom_np[0:2], 22, 0.7), c="green")

    plot_cf(x1_np, 3)
    plot_cf(x2_np, 3)

    ax.plot(x1_np[0], x1_np[1], 'wo', markersize=14, mew=3, zorder=12)
    ax.plot(x1_np[0], x1_np[1], 'o', color='magenta', markersize=10, mew=3, zorder=12)
    ax.plot(x2_np[0], x2_np[1], 'wo', markersize=14, mew=3, zorder=12)
    ax.plot(x2_np[0], x2_np[1], 'o', color='magenta', markersize=10, mew=3, zorder=12)

    ax.set_aspect("equal", "box")

    plt.xlim(0, 10)
    plt.ylim(0, 10)
    # plt.show()
    plt.pause(0.001)

